import 'package:analyzer/dart/element/element.dart';

import 'change_analyzer.dart';
import 'library_visitor.dart';

class ReflectableSourceWrapper {
  final COLLECTION_IMPORT =
      '''import 'dart:collection' show HashSet, UnmodifiableListView;''';
  final MAPPER_IMPORT =
      '''import 'package:dart_json_mapper/dart_json_mapper.dart' show JsonMapper, JsonMapperAdapter, typeOf;''';
  final REFLECTABLE_INIT_METHOD = 'initializeReflectable';
  final REFLECTABLE_INIT_METHOD_PATCH = '_initializeReflectable';
  final INIT_METHOD =
      'JsonMapper initializeJsonMapper({Iterable<JsonMapperAdapter> adapters = const []}) {';

  LibraryVisitor _libraryVisitor = LibraryVisitor();

  LibraryElement inputLibrary;
  Map<String, dynamic> options;
  String lastOutput;

  String _inputLibraryPath;

  ReflectableSourceWrapper(this.inputLibrary, this.options) {
    inputLibrary.visitChildren(_libraryVisitor);
    _inputLibraryPath = inputLibrary.identifier
        .substring(0, inputLibrary.identifier.lastIndexOf('/') + 1);
  }

  Iterable<String> get allowedIterables {
    return (options['iterables'] as String).split(',').map((x) => x.trim());
  }

  bool get isCollectionImportNeeded {
    return allowedIterables.contains('HashSet') ||
        allowedIterables.contains('UnmodifiableListView');
  }

  String get _libraryAdapterId {
    return '$_libraryName';
  }

  String get _libraryName {
    return inputLibrary.identifier
            .split('/')
            .last
            .replaceAll('.dart', '')
            .replaceAll('.', '_') +
        'Adapter';
  }

  String _renderValueDecoratorsForClassElement(ClassElement element) {
    return [
      ...[
        'List',
        'Set'
      ].where((x) => allowedIterables.contains(x)).map((iterable) =>
          '''    typeOf<$iterable<${element.name}>>(): (value) => value.cast<${element.name}>()'''),
      ...[
        'HashSet'
      ].where((x) => allowedIterables.contains(x)).map((iterable) =>
          '''    typeOf<$iterable<${element.name}>>(): (value) => $iterable<${element.name}>.of(value.cast<${element.name}>())'''),
      ...[
        'UnmodifiableListView'
      ].where((x) => allowedIterables.contains(x)).map((iterable) =>
          '''    typeOf<$iterable<${element.name}>>(): (value) => $iterable<${element.name}>(value.cast<${element.name}>())''')
    ].join(',\n');
  }

  String _renderValueDecorators() {
    return _libraryVisitor.visitedPublicAnnotatedClassElements.values
        .map((e) => _renderValueDecoratorsForClassElement(e))
        .join(',\n');
  }

  String _renderLibraryAdapterDefinition() {
    return ''' 
final $_libraryAdapterId = JsonMapperAdapter(
  title: '$_libraryName',
  url: '${inputLibrary.identifier}',
  valueDecorators: {
${_renderValueDecorators()}
});\n''';
  }

  String _renderLibraryAdapterRegistration() {
    return '''
  $REFLECTABLE_INIT_METHOD_PATCH();
  [...adapters, $_libraryAdapterId].forEach((x) => JsonMapper().useAdapter(x));
  return JsonMapper();
}''';
  }

  String _renderElementImport(ClassElement element) {
    if (element.library != null &&
        element.library.identifier.startsWith(_inputLibraryPath)) {
      // local import
      return '''import '${element.library.identifier.split(_inputLibraryPath).last}';''';
    }

    return null;
  }

  String _renderHeader() {
    return '''
// This file has been generated by the dart_json_mapper package.
// https://github.com/k-paxian/dart-json-mapper
''';
  }

  String _renderImports() {
    final importsList = {
      isCollectionImportNeeded ? COLLECTION_IMPORT : null,
      MAPPER_IMPORT,
      ..._libraryVisitor.visitedPublicAnnotatedClassElements.values
          .map((e) => _renderElementImport(e))
    }.where((x) => x != null).toList();
    importsList.sort();
    return importsList.join('\n') + '\n\n';
  }

  String _removeObjectCasts(String input) {
    return input.replaceAll('<Object>', '');
  }

  String _patchInitMethod(String input) {
    final PATCH = '\n' +
        _renderLibraryAdapterDefinition() +
        '\n' +
        INIT_METHOD +
        '\n' +
        _renderLibraryAdapterRegistration();
    return input.replaceFirst(
            REFLECTABLE_INIT_METHOD, REFLECTABLE_INIT_METHOD_PATCH) +
        PATCH;
  }

  bool hasNoIncrementalChanges(LibraryElement library) {
    final incrementalLibraryVisitor = LibraryVisitor();
    library.visitChildren(incrementalLibraryVisitor);
    final hasChanges =
        ChangeAnalyzer(incrementalLibraryVisitor, _libraryVisitor).hasChanges;
    if (hasChanges) {
      _libraryVisitor = incrementalLibraryVisitor;
      inputLibrary = library;
    }
    return !hasChanges;
  }

  String wrap(String reflectableGeneratedSource) {
    lastOutput = _renderHeader() +
        _renderImports() +
        _patchInitMethod(_removeObjectCasts(reflectableGeneratedSource));
    return lastOutput;
  }
}
